﻿using System;

namespace HeuristicLab.Problems.PushGP.ProgramSynthesis {
  using System.Collections.Generic;
  using System.Linq;

  using HeuristicLab.Algorithms.PushGP.Expressions;
  using HeuristicLab.Algorithms.PushGP.Interpreter;
  using HeuristicLab.BenchmarkSuite.ProblemData;
  using HeuristicLab.Common;
  using HeuristicLab.Core;
  using HeuristicLab.Data;
  using HeuristicLab.Encodings.IntegerVectorEncoding;
  using HeuristicLab.Optimization;
  using HeuristicLab.Parameters;
  //using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.Instances;

  //[StorableClass]
  [Creatable(CreatableAttribute.Categories.GeneticProgrammingProblems, Priority = 180)]
  [Item("Push GP Problem", "")]
  public class PushProblem : SingleObjectiveBasicProblem<IntegerVectorEncoding>, IProblemInstanceConsumer<IPushData> {

    private readonly PushGpInterpreterPool pool = new PushGpInterpreterPool();
    private IPushData pushData;

    public PushProblem() {
      this.InitParameters();

      this.Encoding.Bounds[0, 0] = 0;
      this.Encoding.Bounds[0, 1] = ExpressionTable.Count - 1;
      this.Encoding.Length = this.pool.Configuration.MaxPointsInProgram;
    }

    #region Parameters

    private const string EvalPushLimitParameterName = "EvalPushLimit";
    private const string MaxProgramLengthParameterName = "MaxProgramLength";
    private const string TopLevelPushCodeParameterName = "TopLevelPushCode";
    private const string TopLevelPopCodeParameterName = "TopLevelPopCode";

    private void InitParameters() {
      this.Parameters.Add(new FixedValueParameter<IntValue>(EvalPushLimitParameterName, new IntValue(this.pool.Configuration.EvalPushLimit)));
      this.Parameters.Add(new FixedValueParameter<IntValue>(MaxProgramLengthParameterName, new IntValue(this.pool.Configuration.MaxPointsInProgram)));
      this.Parameters.Add(new FixedValueParameter<BoolValue>(TopLevelPushCodeParameterName, new BoolValue(this.pool.Configuration.TopLevelPushCode)) { Hidden = true });
      this.Parameters.Add(new FixedValueParameter<BoolValue>(TopLevelPopCodeParameterName, new BoolValue(this.pool.Configuration.TopLevelPopCode)) { Hidden = true });
    }

    public IValueParameter<IntValue> EvalPushLimitParameter { get { return this.Parameters[EvalPushLimitParameterName] as ValueParameter<IntValue>; } }
    public int EvalPushLimit
    {
      get { return this.pool.Configuration.EvalPushLimit; }
      set
      {
        this.EvalPushLimitParameter.Value.Value = value;
        this.pool.Configuration.EvalPushLimit = value;
      }
    }

    public IValueParameter<IntValue> MaxProgramLengthParameter { get { return this.Parameters[MaxProgramLengthParameterName] as ValueParameter<IntValue>; } }

    public int MaxProgramLength
    {
      get { return this.pool.Configuration.MaxPointsInProgram; }
      set
      {
        this.MaxProgramLengthParameter.Value.Value = value;
        this.Encoding.LengthParameter.Value.Value = value;
        this.pool.Configuration.MaxPointsInProgram = value;
      }
    }

    public IValueParameter<BoolValue> TopLevelPushParameter { get { return this.Parameters[TopLevelPushCodeParameterName] as ValueParameter<BoolValue>; } }

    public bool TopLevelPushCode
    {
      get { return this.pool.Configuration.TopLevelPushCode; }
      set
      {
        this.TopLevelPushParameter.Value.Value = value;
        this.pool.Configuration.TopLevelPushCode = value;
      }
    }

    public IValueParameter<BoolValue> TopLevelPopParameter { get { return this.Parameters[TopLevelPopCodeParameterName] as ValueParameter<BoolValue>; } }

    public bool TopLevelPopCode
    {
      get { return this.pool.Configuration.TopLevelPopCode; }
      set
      {
        this.TopLevelPushParameter.Value.Value = value;
        this.pool.Configuration.TopLevelPopCode = value;
      }
    }

    #endregion

    // persistence
    //[StorableConstructor]
    public PushProblem(bool deserializing) : base(deserializing) { }

    // cloning
    public PushProblem(PushProblem original, Cloner cloner)
      : base(original, cloner) {
    }

    public override bool Maximization { get { return false; } }

    public override IDeepCloneable Clone(Cloner cloner) {
      throw new NotImplementedException();
    }

    public override double Evaluate(Individual individual, IRandom random) {
      var vector = individual.IntegerVector();
      var program = MapToProgram(vector);
      var results = new List<long>();

      using (var interpreter = pool.GetInstance()) {
        for (var i = 0; i < this.pushData.TrainingCount; i++) {
          this.pushData.ApplyTrainingInput(interpreter, i);
          interpreter.Run(program);

          var diff = (long)this.pushData.CompareTrainingOutput(interpreter, i);
          results.Add(diff);

          interpreter.Clear();
        }
      }

      return results.Count == 0 ? 0d : results.Average();
    }

    private static ExecExpandExpression MapToProgram(IntegerVector vector) {
      var expressions = vector.Select(ExpressionTable.GetExpression);

      return new ExecExpandExpression(expressions);
    }

    public void Load(IPushData data) {
      this.pushData = data;
    }
  }
}
